@@ -24,26 +24,73 @@ import (
 	"github.com/evanw/esbuild/pkg/cli"
 )
 
-type responseCallback = func(interface{})
-type rebuildCallback = func(uint32) []byte
-type watchStopCallback = func()
-type serverStopCallback = func()
+type responseCallback func(interface{})
+type rebuildCallback func(uint32) []byte
+type watchStopCallback func()
+type serveStopCallback func()
+type pluginResolveCallback func(uint32, map[string]interface{}) []byte
+
+type activeBuild struct {
+	mutex         sync.Mutex
+	refCount      int
+	rebuild       rebuildCallback
+	watchStop     watchStopCallback
+	serveStop     serveStopCallback
+	pluginResolve pluginResolveCallback
+}
 
 type serviceType struct {
-	mutex           sync.Mutex
-	callbacks       map[uint32]responseCallback
-	rebuilds        map[int]rebuildCallback
-	watchStops      map[int]watchStopCallback
-	serveStops      map[int]serverStopCallback
-	nextID          uint32
-	nextRebuildID   int
-	nextWatchID     int
-	outgoingPackets chan outgoingPacket
+	mutex              sync.Mutex
+	callbacks          map[uint32]responseCallback
+	activeBuilds       map[int]*activeBuild
+	nextRequestID      uint32
+	outgoingPackets    chan outgoingPacket
+	keepAliveWaitGroup sync.WaitGroup
+}
+
+func (service *serviceType) getActiveBuild(key int) *activeBuild {
+	service.mutex.Lock()
+	activeBuild := service.activeBuilds[key]
+	service.mutex.Unlock()
+	return activeBuild
+}
+
+func (service *serviceType) trackActiveBuild(key int, activeBuild *activeBuild) {
+	if activeBuild.refCount > 0 {
+		service.mutex.Lock()
+		defer service.mutex.Unlock()
+		if service.activeBuilds[key] != nil {
+			panic("Internal error")
+		}
+		service.activeBuilds[key] = activeBuild
+
+		// This pairs with "Done()" in "decRefCount"
+		service.keepAliveWaitGroup.Add(1)
+	}
+}
+
+func (service *serviceType) decRefCount(key int, activeBuild *activeBuild) {
+	activeBuild.mutex.Lock()
+	activeBuild.refCount--
+	remaining := activeBuild.refCount
+	activeBuild.mutex.Unlock()
+
+	if remaining < 0 {
+		panic("Internal error")
+	}
+
+	if remaining == 0 {
+		service.mutex.Lock()
+		delete(service.activeBuilds, key)
+		service.mutex.Unlock()
+
+		// This pairs with "Add()" in "trackActiveBuild"
+		service.keepAliveWaitGroup.Done()
+	}
 }
 
 type outgoingPacket struct {
-	bytes    []byte
-	refCount int
+	bytes []byte
 }
 
 func runService(sendPings bool) {
@@ -51,16 +98,13 @@ func runService(sendPings bool) {
 
 	service := serviceType{
 		callbacks:       make(map[uint32]responseCallback),
-		rebuilds:        make(map[int]rebuildCallback),
-		watchStops:      make(map[int]watchStopCallback),
-		serveStops:      make(map[int]serverStopCallback),
+		activeBuilds:    make(map[int]*activeBuild),
 		outgoingPackets: make(chan outgoingPacket),
 	}
 	buffer := make([]byte, 16*1024)
 	stream := []byte{}
 
 	// Write packets on a single goroutine so they aren't interleaved
-	waitGroup := &sync.WaitGroup{}
 	go func() {
 		for {
 			packet, ok := <-service.outgoingPackets
@@ -70,11 +114,7 @@ func runService(sendPings bool) {
 			if _, err := os.Stdout.Write(packet.bytes); err != nil {
 				os.Exit(1) // I/O error
 			}
-
-			// Only signal that this request is done when it has actually been written
-			if packet.refCount != 0 {
-				waitGroup.Add(packet.refCount)
-			}
+			service.keepAliveWaitGroup.Done() // This pairs with the "Add()" when putting stuff into "outgoingPackets"
 		}
 	}()
 
@@ -118,15 +158,14 @@ func runService(sendPings bool) {
 
 			// Clone the input and run it on another goroutine
 			clone := append([]byte{}, packet...)
-			waitGroup.Add(1)
+			service.keepAliveWaitGroup.Add(1)
 			go func() {
 				out := service.handleIncomingPacket(clone)
-				out.refCount--
 				if out.bytes != nil {
+					service.keepAliveWaitGroup.Add(1) // The writer thread will call "Done()"
 					service.outgoingPackets <- out
-				} else if out.refCount != 0 {
-					waitGroup.Add(out.refCount)
 				}
+				service.keepAliveWaitGroup.Done() // This pairs with the "Add()" in the stdin thread
 			}()
 		}
 
@@ -135,7 +174,7 @@ func runService(sendPings bool) {
 	}
 
 	// Wait for the last response to be written to stdout
-	waitGroup.Wait()
+	service.keepAliveWaitGroup.Wait()
 }
 
 func (service *serviceType) sendRequest(request interface{}) interface{} {
@@ -148,11 +187,12 @@ func (service *serviceType) sendRequest(request interface{}) interface{} {
 	id = func() uint32 {
 		service.mutex.Lock()
 		defer service.mutex.Unlock()
-		id := service.nextID
-		service.nextID++
+		id := service.nextRequestID
+		service.nextRequestID++
 		service.callbacks[id] = callback
 		return id
 	}()
+	service.keepAliveWaitGroup.Add(1) // The writer thread will call "Done()"
 	service.outgoingPackets <- outgoingPacket{
 		bytes: encodePacket(packet{
 			id:        id,
@@ -196,100 +236,107 @@ func (service *serviceType) handleIncomingPacket(bytes []byte) (result outgoingP
 				bytes: service.handleTransformRequest(p.id, request),
 			}
 
-		case "rebuild":
-			rebuildID := request["rebuildID"].(int)
-			rebuild, ok := func() (rebuildCallback, bool) {
-				service.mutex.Lock()
-				defer service.mutex.Unlock()
-				rebuild, ok := service.rebuilds[rebuildID]
-				return rebuild, ok
-			}()
-			if !ok {
-				return outgoingPacket{
-					bytes: encodePacket(packet{
-						id: p.id,
-						value: map[string]interface{}{
-							"error": "Cannot rebuild",
-						},
-					}),
+		case "resolve":
+			key := request["key"].(int)
+			if build := service.getActiveBuild(key); build != nil {
+				build.mutex.Lock()
+				pluginResolve := build.pluginResolve
+				build.mutex.Unlock()
+				if pluginResolve != nil {
+					return outgoingPacket{
+						bytes: pluginResolve(p.id, request),
+					}
 				}
 			}
 			return outgoingPacket{
-				bytes: rebuild(p.id),
+				bytes: encodePacket(packet{
+					id: p.id,
+					value: map[string]interface{}{
+						"error": "Cannot call \"resolve\" on an inactive build",
+					},
+				}),
+			}
+
+		case "rebuild":
+			key := request["key"].(int)
+			if build := service.getActiveBuild(key); build != nil {
+				build.mutex.Lock()
+				rebuild := build.rebuild
+				build.mutex.Unlock()
+				if rebuild != nil {
+					return outgoingPacket{
+						bytes: rebuild(p.id),
+					}
+				}
+			}
+			return outgoingPacket{
+				bytes: encodePacket(packet{
+					id: p.id,
+					value: map[string]interface{}{
+						"error": "Cannot rebuild",
+					},
+				}),
 			}
 
 		case "watch-stop":
-			watchID := request["watchID"].(int)
-			refCount := 0
-			watchStop := func() watchStopCallback {
-				// Only mutate the map while inside a mutex
-				service.mutex.Lock()
-				defer service.mutex.Unlock()
-				if watchStop, ok := service.watchStops[watchID]; ok {
-					// This watch is now considered finished. This matches the +1 reference
-					// count at the return of the build call.
-					refCount = -1
-					return watchStop
+			key := request["key"].(int)
+			if build := service.getActiveBuild(key); build != nil {
+				build.mutex.Lock()
+				watchStop := build.watchStop
+				build.watchStop = nil
+				build.mutex.Unlock()
+
+				// Stop watch mode and release this ref count if it was held
+				if watchStop != nil {
+					watchStop()
+					service.decRefCount(key, build)
 				}
-				return nil
-			}()
-			if watchStop != nil {
-				watchStop()
 			}
 			return outgoingPacket{
 				bytes: encodePacket(packet{
 					id:    p.id,
 					value: make(map[string]interface{}),
 				}),
-				refCount: refCount,
 			}
 
 		case "serve-stop":
-			serveID := request["serveID"].(int)
-			refCount := 0
-			serveStop := func() serverStopCallback {
-				// Only mutate the map while inside a mutex
-				service.mutex.Lock()
-				defer service.mutex.Unlock()
-				if serveStop, ok := service.serveStops[serveID]; ok {
-					// This serve is now considered finished. This matches the +1 reference
-					// count at the return of the serve call.
-					refCount = -1
-					return serveStop
+			key := request["key"].(int)
+			if build := service.getActiveBuild(key); build != nil {
+				build.mutex.Lock()
+				serveStop := build.serveStop
+				build.serveStop = nil
+				build.mutex.Unlock()
+
+				// Stop serving but do not release the ref count (that's done elsewhere)
+				if serveStop != nil {
+					serveStop()
 				}
-				return nil
-			}()
-			if serveStop != nil {
-				serveStop()
 			}
 			return outgoingPacket{
 				bytes: encodePacket(packet{
 					id:    p.id,
 					value: make(map[string]interface{}),
 				}),
-				refCount: refCount,
 			}
 
 		case "rebuild-dispose":
-			rebuildID := request["rebuildID"].(int)
-			refCount := 0
-			func() {
-				// Only mutate the map while inside a mutex
-				service.mutex.Lock()
-				defer service.mutex.Unlock()
-				if _, ok := service.rebuilds[rebuildID]; ok {
-					// This build is now considered finished. This matches the +1 reference
-					// count at the return of the first build call for this rebuild chain.
-					refCount = -1
-					delete(service.rebuilds, rebuildID)
+			key := request["key"].(int)
+			if build := service.getActiveBuild(key); build != nil {
+				build.mutex.Lock()
+				rebuild := build.rebuild
+				build.rebuild = nil
+				build.mutex.Unlock()
+
+				// Release this ref count if it was held
+				if rebuild != nil {
+					service.decRefCount(key, build)
 				}
-			}()
+			}
 			return outgoingPacket{
 				bytes: encodePacket(packet{
 					id:    p.id,
 					value: make(map[string]interface{}),
 				}),
-				refCount: refCount,
 			}
 
 		case "error":
@@ -397,8 +444,12 @@ func (service *serviceType) handleBuildRequest(id uint32, request map[string]int
 		}
 	}
 
+	activeBuild := &activeBuild{refCount: 1}
+	service.trackActiveBuild(key, activeBuild)
+	defer service.decRefCount(key, activeBuild)
+
 	if plugins, ok := request["plugins"]; ok {
-		if plugins, err := service.convertPlugins(key, plugins); err != nil {
+		if plugins, err := service.convertPlugins(key, plugins, activeBuild); err != nil {
 			return outgoingPacket{bytes: encodeErrorPacket(id, err)}
 		} else {
 			options.Plugins = plugins
@@ -406,33 +457,20 @@ func (service *serviceType) handleBuildRequest(id uint32, request map[string]int
 	}
 
 	if isServe {
-		return service.handleServeRequest(id, options, serveObj)
-	}
-
-	rebuildID := service.nextRebuildID
-	watchID := service.nextWatchID
-	if incremental {
-		service.nextRebuildID++
-	}
-	if options.Watch != nil {
-		service.nextWatchID++
+		return service.handleServeRequest(id, options, serveObj, key, activeBuild)
 	}
 
 	resultToResponse := func(result api.BuildResult) map[string]interface{} {
 		response := map[string]interface{}{
 			"errors":   encodeMessages(result.Errors),
 			"warnings": encodeMessages(result.Warnings),
+			"rebuild":  incremental,
+			"watch":    options.Watch != nil,
 		}
 		if !write {
 			// Pass the output files back to the caller
 			response["outputFiles"] = encodeOutputFiles(result.OutputFiles)
 		}
-		if incremental {
-			response["rebuildID"] = rebuildID
-		}
-		if options.Watch != nil {
-			response["watchID"] = watchID
-		}
 		if options.Metafile {
 			response["metafile"] = result.Metafile
 		}
@@ -446,7 +484,7 @@ func (service *serviceType) handleBuildRequest(id uint32, request map[string]int
 		options.Watch.OnRebuild = func(result api.BuildResult) {
 			service.sendRequest(map[string]interface{}{
 				"command": "watch-rebuild",
-				"watchID": watchID,
+				"key":     key,
 				"args":    resultToResponse(result),
 			})
 		}
@@ -458,39 +496,28 @@ func (service *serviceType) handleBuildRequest(id uint32, request map[string]int
 	options.Incremental = incremental
 	result := api.Build(options)
 	response := resultToResponse(result)
-	refCount := 0
 
 	if incremental {
-		func() {
-			// Only mutate the map while inside a mutex
-			service.mutex.Lock()
-			defer service.mutex.Unlock()
-			service.rebuilds[rebuildID] = func(id uint32) []byte {
-				result := result.Rebuild()
-				response := resultToResponse(result)
-				return encodePacket(packet{
-					id:    id,
-					value: response,
-				})
-			}
-		}()
+		activeBuild.rebuild = func(id uint32) []byte {
+			result := result.Rebuild()
+			response := resultToResponse(result)
+			return encodePacket(packet{
+				id:    id,
+				value: response,
+			})
+		}
 
 		// Make sure the build doesn't finish until "dispose" has been called
-		refCount++
+		activeBuild.refCount++
 	}
 
 	if options.Watch != nil {
-		func() {
-			// Only mutate the map while inside a mutex
-			service.mutex.Lock()
-			defer service.mutex.Unlock()
-			service.watchStops[watchID] = func() {
-				result.Stop()
-			}
-		}()
+		activeBuild.watchStop = func() {
+			result.Stop()
+		}
 
 		// Make sure the build doesn't finish until "stop" has been called
-		refCount++
+		activeBuild.refCount++
 	}
 
 	return outgoingPacket{
@@ -498,14 +525,12 @@ func (service *serviceType) handleBuildRequest(id uint32, request map[string]int
 			id:    id,
 			value: response,
 		}),
-		refCount: refCount,
 	}
 }
 
-func (service *serviceType) handleServeRequest(id uint32, options api.BuildOptions, serveObj interface{}) outgoingPacket {
+func (service *serviceType) handleServeRequest(id uint32, options api.BuildOptions, serveObj interface{}, key int, activeBuild *activeBuild) outgoingPacket {
 	var serveOptions api.ServeOptions
 	serve := serveObj.(map[string]interface{})
-	serveID := serve["serveID"].(int)
 	if port, ok := serve["port"]; ok {
 		serveOptions.Port = uint16(port.(int))
 	}
@@ -518,7 +543,7 @@ func (service *serviceType) handleServeRequest(id uint32, options api.BuildOptio
 	serveOptions.OnRequest = func(args api.ServeOnRequestArgs) {
 		service.sendRequest(map[string]interface{}{
 			"command": "serve-request",
-			"serveID": serveID,
+			"key":     key,
 			"args": map[string]interface{}{
 				"remoteAddress": args.RemoteAddress,
 				"method":        args.Method,
@@ -537,11 +562,14 @@ func (service *serviceType) handleServeRequest(id uint32, options api.BuildOptio
 		"host": result.Host,
 	}
 
+	activeBuild.refCount++ // Make sure the serve doesn't finish until "Wait" finishes
+	activeBuild.serveStop = result.Stop
+
 	// Asynchronously wait for the server to stop, then fulfil the "wait" promise
 	go func() {
 		request := map[string]interface{}{
 			"command": "serve-wait",
-			"serveID": serveID,
+			"key":     key,
 		}
 		if err := result.Wait(); err != nil {
 			request["error"] = err.Error()
@@ -550,17 +578,8 @@ func (service *serviceType) handleServeRequest(id uint32, options api.BuildOptio
 		}
 		service.sendRequest(request)
 
-		// Only mutate the map while inside a mutex
-		service.mutex.Lock()
-		defer service.mutex.Unlock()
-		delete(service.serveStops, serveID)
-	}()
-
-	func() {
-		// Only mutate the map while inside a mutex
-		service.mutex.Lock()
-		defer service.mutex.Unlock()
-		service.serveStops[serveID] = result.Stop
+		// Release the ref count now that the server has shut down
+		service.decRefCount(key, activeBuild)
 	}()
 
 	return outgoingPacket{
@@ -568,15 +587,61 @@ func (service *serviceType) handleServeRequest(id uint32, options api.BuildOptio
 			id:    id,
 			value: response,
 		}),
-
-		// Make sure the serve doesn't finish until "stop" has been called
-		refCount: 1,
 	}
 }
 
-func (service *serviceType) convertPlugins(key int, jsPlugins interface{}) ([]api.Plugin, error) {
-	var goPlugins []api.Plugin
+func resolveKindToString(kind api.ResolveKind) string {
+	switch kind {
+	case api.ResolveEntryPoint:
+		return "entry-point"
 
+	// JS
+	case api.ResolveJSImportStatement:
+		return "import-statement"
+	case api.ResolveJSRequireCall:
+		return "require-call"
+	case api.ResolveJSDynamicImport:
+		return "dynamic-import"
+	case api.ResolveJSRequireResolve:
+		return "require-resolve"
+
+	// CSS
+	case api.ResolveCSSImportRule:
+		return "import-rule"
+	case api.ResolveCSSURLToken:
+		return "url-token"
+
+	default:
+		panic("Internal error")
+	}
+}
+
+func stringToResolveKind(kind string) (api.ResolveKind, bool) {
+	switch kind {
+	case "entry-point":
+		return api.ResolveEntryPoint, true
+
+	// JS
+	case "import-statement":
+		return api.ResolveJSImportStatement, true
+	case "require-call":
+		return api.ResolveJSRequireCall, true
+	case "dynamic-import":
+		return api.ResolveJSDynamicImport, true
+	case "require-resolve":
+		return api.ResolveJSRequireResolve, true
+
+	// CSS
+	case "import-rule":
+		return api.ResolveCSSImportRule, true
+	case "url-token":
+		return api.ResolveCSSURLToken, true
+	}
+
+	return api.ResolveEntryPoint, false
+}
+
+func (service *serviceType) convertPlugins(key int, jsPlugins interface{}, activeBuild *activeBuild) ([]api.Plugin, error) {
 	type filteredCallback struct {
 		filter     *regexp.Regexp
 		pluginName string
@@ -624,14 +689,64 @@ func (service *serviceType) convertPlugins(key int, jsPlugins interface{}) ([]ap
 	// We want to minimize the amount of IPC traffic. Instead of adding one Go
 	// plugin for every JavaScript plugin, we just add a single Go plugin that
 	// proxies the plugin queries to the list of JavaScript plugins in the host.
-	goPlugins = append(goPlugins, api.Plugin{
+	return []api.Plugin{{
 		Name: "JavaScript plugins",
 		Setup: func(build api.PluginBuild) {
+			activeBuild.mutex.Lock()
+			activeBuild.pluginResolve = func(id uint32, request map[string]interface{}) []byte {
+				path := request["path"].(string)
+				var options api.ResolveOptions
+				if value, ok := request["pluginName"]; ok {
+					options.PluginName = value.(string)
+				}
+				if value, ok := request["importer"]; ok {
+					options.Importer = value.(string)
+				}
+				if value, ok := request["namespace"]; ok {
+					options.Namespace = value.(string)
+				}
+				if value, ok := request["resolveDir"]; ok {
+					options.ResolveDir = value.(string)
+				}
+				if value, ok := request["kind"]; ok {
+					str := value.(string)
+					kind, ok := stringToResolveKind(str)
+					if !ok {
+						return encodePacket(packet{
+							id: id,
+							value: map[string]interface{}{
+								"error": fmt.Sprintf("Invalid kind: %q", str),
+							},
+						})
+					}
+					options.Kind = kind
+				}
+				if value, ok := request["pluginData"]; ok {
+					options.PluginData = value.(int)
+				}
+
+				result := build.Resolve(path, options)
+				return encodePacket(packet{
+					id: id,
+					value: map[string]interface{}{
+						"errors":      encodeMessages(result.Errors),
+						"warnings":    encodeMessages(result.Warnings),
+						"path":        result.Path,
+						"external":    result.External,
+						"sideEffects": result.SideEffects,
+						"namespace":   result.Namespace,
+						"suffix":      result.Suffix,
+						"pluginData":  result.PluginData,
+					},
+				})
+			}
+			activeBuild.mutex.Unlock()
+
 			build.OnStart(func() (api.OnStartResult, error) {
 				result := api.OnStartResult{}
 
 				response := service.sendRequest(map[string]interface{}{
-					"command": "start",
+					"command": "on-start",
 					"key":     key,
 				}).(map[string]interface{})
 
@@ -659,40 +774,15 @@ func (service *serviceType) convertPlugins(key int, jsPlugins interface{}) ([]ap
 					return result, nil
 				}
 
-				var kind string
-				switch args.Kind {
-				case api.ResolveEntryPoint:
-					kind = "entry-point"
-
-				// JS
-				case api.ResolveJSImportStatement:
-					kind = "import-statement"
-				case api.ResolveJSRequireCall:
-					kind = "require-call"
-				case api.ResolveJSDynamicImport:
-					kind = "dynamic-import"
-				case api.ResolveJSRequireResolve:
-					kind = "require-resolve"
-
-				// CSS
-				case api.ResolveCSSImportRule:
-					kind = "import-rule"
-				case api.ResolveCSSURLToken:
-					kind = "url-token"
-
-				default:
-					panic("Internal error")
-				}
-
 				response := service.sendRequest(map[string]interface{}{
-					"command":    "resolve",
+					"command":    "on-resolve",
 					"key":        key,
 					"ids":        ids,
 					"path":       args.Path,
 					"importer":   args.Importer,
 					"namespace":  args.Namespace,
 					"resolveDir": args.ResolveDir,
-					"kind":       kind,
+					"kind":       resolveKindToString(args.Kind),
 					"pluginData": args.PluginData,
 				}).(map[string]interface{})
 
@@ -717,6 +807,9 @@ func (service *serviceType) convertPlugins(key int, jsPlugins interface{}) ([]ap
 				if value, ok := response["namespace"]; ok {
 					result.Namespace = value.(string)
 				}
+				if value, ok := response["suffix"]; ok {
+					result.Suffix = value.(string)
+				}
 				if value, ok := response["external"]; ok {
 					result.External = value.(bool)
 				}
@@ -761,11 +854,12 @@ func (service *serviceType) convertPlugins(key int, jsPlugins interface{}) ([]ap
 				}
 
 				response := service.sendRequest(map[string]interface{}{
-					"command":    "load",
+					"command":    "on-load",
 					"key":        key,
 					"ids":        ids,
 					"path":       args.Path,
 					"namespace":  args.Namespace,
+					"suffix":     args.Suffix,
 					"pluginData": args.PluginData,
 				}).(map[string]interface{})
 
@@ -817,9 +911,7 @@ func (service *serviceType) convertPlugins(key int, jsPlugins interface{}) ([]ap
 				return result, nil
 			})
 		},
-	})
-
-	return goPlugins, nil
+	}}, nil
 }
 
 func (service *serviceType) handleTransformRequest(id uint32, request map[string]interface{}) []byte {
