@@ -377,7 +377,7 @@ func parseFile(args parseArgs) {
 
 				// Ignore records that the parser has discarded. This is used to remove
 				// type-only imports in TypeScript files.
-				if record.IsUnused {
+				if record.Flags.Has(ast.IsUnused) {
 					continue
 				}
 
@@ -393,7 +393,7 @@ func parseFile(args parseArgs) {
 				}
 
 				// Run the resolver and log an error if the path couldn't be resolved
-				resolveResult, didLogError, debug := runOnResolvePlugins(
+				resolveResult, didLogError, debug := RunOnResolvePlugins(
 					args.options.Plugins,
 					args.res,
 					args.log,
@@ -401,7 +401,7 @@ func parseFile(args parseArgs) {
 					&args.caches.FSCache,
 					&source,
 					record.Range,
-					source.KeyPath.Namespace,
+					source.KeyPath,
 					record.Path.Text,
 					record.Kind,
 					absResolveDir,
@@ -415,7 +415,7 @@ func parseFile(args parseArgs) {
 					if resolveResult != nil && resolveResult.IsExternal {
 						// Allow path substitution as long as the result is external
 						result.resolveResults[importRecordIndex] = resolveResult
-					} else if !record.HandlesImportErrors {
+					} else if !record.Flags.Has(ast.HandlesImportErrors) {
 						args.log.Add(logger.Warning, &tracker, record.Range,
 							fmt.Sprintf("%q should be marked as external for use with \"require.resolve\"", record.Path.Text))
 					}
@@ -427,48 +427,11 @@ func parseFile(args parseArgs) {
 					// external imports instead of causing errors. This matches a common
 					// code pattern for conditionally importing a module with a graceful
 					// fallback.
-					if !didLogError && !record.HandlesImportErrors {
-						hint := ""
-						if resolver.IsPackagePath(record.Path.Text) {
-							hint = fmt.Sprintf("You can mark the path %q as external to exclude it from the bundle, which will remove this error.", record.Path.Text)
-							if record.Kind == ast.ImportRequire {
-								hint += " You can also surround this \"require\" call with a try/catch block to handle this failure at run-time instead of bundle-time."
-							} else if record.Kind == ast.ImportDynamic {
-								hint += " You can also add \".catch()\" here to handle this failure at run-time instead of bundle-time."
-							}
-							if pluginName == "" && !args.fs.IsAbs(record.Path.Text) {
-								if query := args.res.ProbeResolvePackageAsRelative(absResolveDir, record.Path.Text, record.Kind); query != nil {
-									hint = fmt.Sprintf("Use the relative path %q to reference the file %q. "+
-										"Without the leading \"./\", the path %q is being interpreted as a package path instead.",
-										"./"+record.Path.Text, args.res.PrettyPath(query.PathPair.Primary), record.Path.Text)
-								}
-							}
-						}
-						if args.options.Platform != config.PlatformNode {
-							if _, ok := resolver.BuiltInNodeModules[record.Path.Text]; ok {
-								var how string
-								switch logger.API {
-								case logger.CLIAPI:
-									how = "--platform=node"
-								case logger.JSAPI:
-									how = "platform: 'node'"
-								case logger.GoAPI:
-									how = "Platform: api.PlatformNode"
-								}
-								hint = fmt.Sprintf("The package %q wasn't found on the file system but is built into node. "+
-									"Are you trying to bundle for node? You can use %q to do that, which will remove this error.", record.Path.Text, how)
-							}
-						}
-						if absResolveDir == "" && pluginName != "" {
-							hint = fmt.Sprintf("The plugin %q didn't set a resolve directory for the file %q, "+
-								"so esbuild did not search for %q on the file system.", pluginName, source.PrettyPath, record.Path.Text)
-						}
-						var notes []logger.MsgData
-						if hint != "" {
-							notes = append(notes, logger.MsgData{Text: hint})
-						}
-						debug.LogErrorMsg(args.log, &source, record.Range, fmt.Sprintf("Could not resolve %q", record.Path.Text), notes)
-					} else if args.log.Level <= logger.LevelDebug && !didLogError && record.HandlesImportErrors {
+					if !didLogError && !record.Flags.Has(ast.HandlesImportErrors) {
+						text, notes := ResolveFailureErrorTextAndNotes(args.res, record.Path.Text, record.Kind,
+							pluginName, args.fs, absResolveDir, args.options.Platform, source.PrettyPath)
+						debug.LogErrorMsg(args.log, &source, record.Range, text, notes)
+					} else if args.log.Level <= logger.LevelDebug && !didLogError && record.Flags.Has(ast.HandlesImportErrors) {
 						args.log.Add(logger.Debug, &tracker, record.Range,
 							fmt.Sprintf("Importing %q was allowed even though it could not be resolved because dynamic import failures appear to be handled here:",
 								record.Path.Text))
@@ -505,6 +468,66 @@ func parseFile(args parseArgs) {
 	args.results <- result
 }
 
+func ResolveFailureErrorTextAndNotes(
+	res resolver.Resolver,
+	path string,
+	kind ast.ImportKind,
+	pluginName string,
+	fs fs.FS,
+	absResolveDir string,
+	platform config.Platform,
+	originatingFilePath string,
+) (string, []logger.MsgData) {
+	hint := ""
+
+	if resolver.IsPackagePath(path) {
+		hint = fmt.Sprintf("You can mark the path %q as external to exclude it from the bundle, which will remove this error.", path)
+		if kind == ast.ImportRequire {
+			hint += " You can also surround this \"require\" call with a try/catch block to handle this failure at run-time instead of bundle-time."
+		} else if kind == ast.ImportDynamic {
+			hint += " You can also add \".catch()\" here to handle this failure at run-time instead of bundle-time."
+		}
+		if pluginName == "" && !fs.IsAbs(path) {
+			if query := res.ProbeResolvePackageAsRelative(absResolveDir, path, kind); query != nil {
+				hint = fmt.Sprintf("Use the relative path %q to reference the file %q. "+
+					"Without the leading \"./\", the path %q is being interpreted as a package path instead.",
+					"./"+path, res.PrettyPath(query.PathPair.Primary), path)
+			}
+		}
+	}
+
+	if platform != config.PlatformNode {
+		if _, ok := resolver.BuiltInNodeModules[path]; ok {
+			var how string
+			switch logger.API {
+			case logger.CLIAPI:
+				how = "--platform=node"
+			case logger.JSAPI:
+				how = "platform: 'node'"
+			case logger.GoAPI:
+				how = "Platform: api.PlatformNode"
+			}
+			hint = fmt.Sprintf("The package %q wasn't found on the file system but is built into node. "+
+				"Are you trying to bundle for node? You can use %q to do that, which will remove this error.", path, how)
+		}
+	}
+
+	if absResolveDir == "" && pluginName != "" {
+		where := ""
+		if originatingFilePath != "" {
+			where = fmt.Sprintf(" for the file %q", originatingFilePath)
+		}
+		hint = fmt.Sprintf("The plugin %q didn't set a resolve directory%s, "+
+			"so esbuild did not search for %q on the file system.", pluginName, where, path)
+	}
+
+	var notes []logger.MsgData
+	if hint != "" {
+		notes = append(notes, logger.MsgData{Text: hint})
+	}
+	return fmt.Sprintf("Could not resolve %q", path), notes
+}
+
 func joinWithPublicPath(publicPath string, relPath string) string {
 	if strings.HasPrefix(relPath, "./") {
 		relPath = relPath[2:]
@@ -637,10 +660,10 @@ func logPluginMessages(
 			msg.Data.Location = tracker.MsgLocationOrNil(importPathRange)
 		} else {
 			sanitizeLocation(res, msg.Data.Location)
-			if msg.Data.Location.File == "" && importSource != nil {
-				msg.Data.Location.File = importSource.PrettyPath
-			}
 			if importSource != nil {
+				if msg.Data.Location.File == "" {
+					msg.Data.Location.File = importSource.PrettyPath
+				}
 				msg.Notes = append(msg.Notes, tracker.MsgData(importPathRange,
 					fmt.Sprintf("The plugin %q was triggered by this import", name)))
 			}
@@ -667,7 +690,7 @@ func logPluginMessages(
 	return didLogError
 }
 
-func runOnResolvePlugins(
+func RunOnResolvePlugins(
 	plugins []config.Plugin,
 	res resolver.Resolver,
 	log logger.Log,
@@ -675,7 +698,7 @@ func runOnResolvePlugins(
 	fsCache *cache.FSCache,
 	importSource *logger.Source,
 	importPathRange logger.Range,
-	importNamespace string,
+	importer logger.Path,
 	path string,
 	kind ast.ImportKind,
 	absResolveDir string,
@@ -686,15 +709,11 @@ func runOnResolvePlugins(
 		ResolveDir: absResolveDir,
 		Kind:       kind,
 		PluginData: pluginData,
+		Importer:   importer,
 	}
 	applyPath := logger.Path{
 		Text:      path,
-		Namespace: importNamespace,
-	}
-	if importSource != nil {
-		resolverArgs.Importer = importSource.KeyPath
-	} else {
-		resolverArgs.Importer.Namespace = importNamespace
+		Namespace: importer.Namespace,
 	}
 	tracker := logger.MakeLineColumnTracker(importSource)
 
@@ -1125,9 +1144,9 @@ func (s *scanner) maybeParseFile(
 
 	// Set the module type preference using node's module type rules
 	if strings.HasSuffix(path.Text, ".mjs") || strings.HasSuffix(path.Text, ".mts") {
-		optionsClone.ModuleType = config.ModuleESM
+		optionsClone.ModuleType = js_ast.ModuleESM
 	} else if strings.HasSuffix(path.Text, ".cjs") || strings.HasSuffix(path.Text, ".cts") {
-		optionsClone.ModuleType = config.ModuleCommonJS
+		optionsClone.ModuleType = js_ast.ModuleCommonJS
 	} else {
 		optionsClone.ModuleType = resolveResult.ModuleType
 	}
@@ -1362,13 +1381,13 @@ func (s *scanner) addEntryPoints(entryPoints []EntryPoint) []graph.EntryPoint {
 	entryPointWaitGroup.Add(len(entryPoints))
 	for i, entryPoint := range entryPoints {
 		go func(i int, entryPoint EntryPoint) {
-			namespace := ""
+			var importer logger.Path
 			if entryPoint.IsFile {
-				namespace = "file"
+				importer.Namespace = "file"
 			}
 
 			// Run the resolver and log an error if the path couldn't be resolved
-			resolveResult, didLogError, debug := runOnResolvePlugins(
+			resolveResult, didLogError, debug := RunOnResolvePlugins(
 				s.options.Plugins,
 				s.res,
 				s.log,
@@ -1376,7 +1395,7 @@ func (s *scanner) addEntryPoints(entryPoints []EntryPoint) []graph.EntryPoint {
 				&s.caches.FSCache,
 				nil,
 				logger.Range{},
-				namespace,
+				importer,
 				entryPoint.InputPath,
 				ast.ImportEntryPoint,
 				entryPointAbsResolveDir,
@@ -1584,6 +1603,11 @@ func (s *scanner) scanAllDependencies() {
 						&result.file.inputFile.Source, record.Range, resolveResult.PluginData, inputKindNormal, nil)
 					record.SourceIndex = ast.MakeIndex32(sourceIndex)
 				} else {
+					// Allow this import statement to be removed if something marked it as "sideEffects: false"
+					if resolveResult.PrimarySideEffectsData != nil {
+						record.Flags |= ast.IsExternalWithoutSideEffects
+					}
+
 					// If the path to the external module is relative to the source
 					// file, rewrite the path to be relative to the working directory
 					if path.Namespace == "file" {
@@ -1751,7 +1775,7 @@ func (s *scanner) processScannedFiles() []scannerFile {
 				// about it. Note that this can result in esbuild silently generating
 				// broken code. If this actually happens for people, it's probably worth
 				// re-enabling the warning about code inside "node_modules".
-				if record.WasOriginallyBareImport && !s.options.IgnoreDCEAnnotations &&
+				if record.Flags.Has(ast.WasOriginallyBareImport) && !s.options.IgnoreDCEAnnotations &&
 					!helpers.IsInsideNodeModules(result.file.inputFile.Source.KeyPath.Text) {
 					if otherModule := &s.results[record.SourceIndex.GetIndex()].file.inputFile; otherModule.SideEffects.Kind != graph.HasSideEffects &&
 						// Do not warn if this is from a plugin, since removing the import
