@@ -183,9 +183,9 @@ func TestPackageJsonSideEffectsFalseKeepBareImportAndRequireES6(t *testing.T) {
 			Mode:          config.ModeBundle,
 			AbsOutputFile: "/out.js",
 		},
-		expectedScanLog: `Users/user/project/src/entry.js: warning: Ignoring this import because "` +
+		expectedScanLog: `Users/user/project/src/entry.js: WARNING: Ignoring this import because "` +
 			`Users/user/project/node_modules/demo-pkg/index.js" was marked as having no side effects
-Users/user/project/node_modules/demo-pkg/package.json: note: "sideEffects" is false ` +
+Users/user/project/node_modules/demo-pkg/package.json: NOTE: "sideEffects" is false ` +
 			`in the enclosing "package.json" file
 `,
 	})
@@ -214,9 +214,9 @@ func TestPackageJsonSideEffectsFalseKeepBareImportAndRequireCommonJS(t *testing.
 			Mode:          config.ModeBundle,
 			AbsOutputFile: "/out.js",
 		},
-		expectedScanLog: `Users/user/project/src/entry.js: warning: Ignoring this import because "` +
+		expectedScanLog: `Users/user/project/src/entry.js: WARNING: Ignoring this import because "` +
 			`Users/user/project/node_modules/demo-pkg/index.js" was marked as having no side effects
-Users/user/project/node_modules/demo-pkg/package.json: note: "sideEffects" is false ` +
+Users/user/project/node_modules/demo-pkg/package.json: NOTE: "sideEffects" is false ` +
 			`in the enclosing "package.json" file
 `,
 	})
@@ -244,9 +244,9 @@ func TestPackageJsonSideEffectsFalseRemoveBareImportES6(t *testing.T) {
 			Mode:          config.ModeBundle,
 			AbsOutputFile: "/out.js",
 		},
-		expectedScanLog: `Users/user/project/src/entry.js: warning: Ignoring this import because "` +
+		expectedScanLog: `Users/user/project/src/entry.js: WARNING: Ignoring this import because "` +
 			`Users/user/project/node_modules/demo-pkg/index.js" was marked as having no side effects
-Users/user/project/node_modules/demo-pkg/package.json: note: "sideEffects" is false ` +
+Users/user/project/node_modules/demo-pkg/package.json: NOTE: "sideEffects" is false ` +
 			`in the enclosing "package.json" file
 `,
 	})
@@ -274,9 +274,9 @@ func TestPackageJsonSideEffectsFalseRemoveBareImportCommonJS(t *testing.T) {
 			Mode:          config.ModeBundle,
 			AbsOutputFile: "/out.js",
 		},
-		expectedScanLog: `Users/user/project/src/entry.js: warning: Ignoring this import because "` +
+		expectedScanLog: `Users/user/project/src/entry.js: WARNING: Ignoring this import because "` +
 			`Users/user/project/node_modules/demo-pkg/index.js" was marked as having no side effects
-Users/user/project/node_modules/demo-pkg/package.json: note: "sideEffects" is false ` +
+Users/user/project/node_modules/demo-pkg/package.json: NOTE: "sideEffects" is false ` +
 			`in the enclosing "package.json" file
 `,
 	})
@@ -722,9 +722,9 @@ func TestPackageJsonSideEffectsArrayGlob(t *testing.T) {
 			Mode:          config.ModeBundle,
 			AbsOutputFile: "/out.js",
 		},
-		expectedScanLog: `Users/user/project/src/entry.js: warning: Ignoring this import because ` +
+		expectedScanLog: `Users/user/project/src/entry.js: WARNING: Ignoring this import because ` +
 			`"Users/user/project/node_modules/demo-pkg/remove/this/file.js" was marked as having no side effects
-Users/user/project/node_modules/demo-pkg/package.json: note: It was excluded from the "sideEffects" ` +
+Users/user/project/node_modules/demo-pkg/package.json: NOTE: It was excluded from the "sideEffects" ` +
 			`array in the enclosing "package.json" file
 `,
 	})
@@ -1937,3 +1937,47 @@ func TestDCEClassStaticBlocks(t *testing.T) {
 		},
 	})
 }
+
+func TestDCEVarExports(t *testing.T) {
+	dce_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/a.js": `
+				var foo = { bar: 123 }
+				module.exports = foo
+			`,
+			"/b.js": `
+				var exports = { bar: 123 }
+				module.exports = exports
+			`,
+			"/c.js": `
+				var module = { bar: 123 }
+				exports.foo = module
+			`,
+		},
+		entryPaths: []string{"/a.js", "/b.js", "/c.js"},
+		options: config.Options{
+			Mode:         config.ModeBundle,
+			AbsOutputDir: "/out",
+		},
+	})
+}
+
+func TestDCETemplateLiteral(t *testing.T) {
+	dce_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.js": "" +
+				"var remove;\n" +
+				"var alsoKeep;\n" +
+				"let a = `${keep}`\n" +
+				"let b = `${123}`\n" +
+				"let c = `${keep ? 1 : 2n}`\n" +
+				"let d = `${remove ? 1 : 2n}`\n" +
+				"let e = `${alsoKeep}`\n",
+		},
+		entryPaths: []string{"/entry.js"},
+		options: config.Options{
+			Mode:         config.ModeBundle,
+			AbsOutputDir: "/out",
+		},
+	})
+}
