@@ -460,7 +460,9 @@ type ENewTarget struct {
 	Range logger.Range
 }
 
-type EImportMeta struct{}
+type EImportMeta struct {
+	RangeLen int32
+}
 
 // These help reduce unnecessary memory allocations
 var BMissingShared = &BMissing{}
@@ -469,7 +471,6 @@ var ESuperShared = &ESuper{}
 var ENullShared = &ENull{}
 var EUndefinedShared = &EUndefined{}
 var EThisShared = &EThis{}
-var EImportMetaShared = &EImportMeta{}
 
 type ENew struct {
 	Target Expr
@@ -776,7 +777,7 @@ func MaybeSimplifyNot(expr Expr) (Expr, bool) {
 
 	case *EUnary:
 		// "!!!a" => "!a"
-		if e.Op == UnOpNot && IsBooleanValue(e.Value) {
+		if e.Op == UnOpNot && KnownPrimitiveType(e.Value) == PrimitiveBoolean {
 			return e.Value, true
 		}
 
@@ -815,109 +816,158 @@ func MaybeSimplifyNot(expr Expr) (Expr, bool) {
 	return Expr{}, false
 }
 
-func IsBooleanValue(a Expr) bool {
+type PrimitiveType uint8
+
+const (
+	PrimitiveUnknown PrimitiveType = iota
+	PrimitiveMixed
+	PrimitiveNull
+	PrimitiveUndefined
+	PrimitiveBoolean
+	PrimitiveNumber
+	PrimitiveString
+	PrimitiveBigInt
+)
+
+// This can be used when the returned type is either one or the other
+func MergedKnownPrimitiveTypes(a Expr, b Expr) PrimitiveType {
+	x := KnownPrimitiveType(a)
+	y := KnownPrimitiveType(b)
+	if x == PrimitiveUnknown || y == PrimitiveUnknown {
+		return PrimitiveUnknown
+	}
+	if x == y {
+		return x
+	}
+	return PrimitiveMixed // Definitely some kind of primitive
+}
+
+func KnownPrimitiveType(a Expr) PrimitiveType {
 	switch e := a.Data.(type) {
 	case *EInlinedEnum:
-		return IsBooleanValue(e.Value)
+		return KnownPrimitiveType(e.Value)
+
+	case *ENull:
+		return PrimitiveNull
+
+	case *EUndefined:
+		return PrimitiveUndefined
 
 	case *EBoolean:
-		return true
+		return PrimitiveBoolean
+
+	case *ENumber:
+		return PrimitiveNumber
+
+	case *EString:
+		return PrimitiveString
+
+	case *EBigInt:
+		return PrimitiveBigInt
+
+	case *ETemplate:
+		if e.TagOrNil.Data == nil {
+			return PrimitiveString
+		}
 
 	case *EIf:
-		return IsBooleanValue(e.Yes) && IsBooleanValue(e.No)
+		return MergedKnownPrimitiveTypes(e.Yes, e.No)
 
 	case *EUnary:
-		return e.Op == UnOpNot || e.Op == UnOpDelete
+		switch e.Op {
+		case UnOpVoid:
+			return PrimitiveUndefined
+
+		case UnOpTypeof:
+			return PrimitiveString
+
+		case UnOpNot, UnOpDelete:
+			return PrimitiveBoolean
+
+		case UnOpPos:
+			return PrimitiveNumber // Cannot be bigint because that throws an exception
+
+		case UnOpNeg, UnOpCpl:
+			value := KnownPrimitiveType(e.Value)
+			if value == PrimitiveBigInt {
+				return PrimitiveBigInt
+			}
+			if value != PrimitiveUnknown && value != PrimitiveMixed {
+				return PrimitiveNumber
+			}
+			return PrimitiveMixed // Can be number or bigint
+
+		case UnOpPreDec, UnOpPreInc, UnOpPostDec, UnOpPostInc:
+			return PrimitiveMixed // Can be number or bigint
+		}
 
 	case *EBinary:
 		switch e.Op {
 		case BinOpStrictEq, BinOpStrictNe, BinOpLooseEq, BinOpLooseNe,
 			BinOpLt, BinOpGt, BinOpLe, BinOpGe,
 			BinOpInstanceof, BinOpIn:
-			return true
+			return PrimitiveBoolean
 
 		case BinOpLogicalOr, BinOpLogicalAnd:
-			return IsBooleanValue(e.Left) && IsBooleanValue(e.Right)
+			return MergedKnownPrimitiveTypes(e.Left, e.Right)
 
 		case BinOpNullishCoalescing:
-			return IsBooleanValue(e.Left)
-		}
-	}
+			left := KnownPrimitiveType(e.Left)
+			right := KnownPrimitiveType(e.Right)
+			if left == PrimitiveNull || left == PrimitiveUndefined {
+				return right
+			}
+			if left != PrimitiveUnknown {
+				if left != PrimitiveMixed {
+					return left // Definitely not null or undefined
+				}
+				if right != PrimitiveUnknown {
+					return PrimitiveMixed // Definitely some kind of primitive
+				}
+			}
 
-	return false
-}
-
-func IsNumericValue(a Expr) bool {
-	switch e := a.Data.(type) {
-	case *EInlinedEnum:
-		return IsNumericValue(e.Value)
-
-	case *ENumber:
-		return true
-
-	case *EIf:
-		return IsNumericValue(e.Yes) && IsNumericValue(e.No)
-
-	case *EUnary:
-		switch e.Op {
-		case UnOpPos, UnOpNeg, UnOpCpl, UnOpPreDec, UnOpPreInc, UnOpPostDec, UnOpPostInc:
-			return true
-		}
-
-	case *EBinary:
-		switch e.Op {
 		case BinOpAdd:
-			return IsNumericValue(e.Left) && IsNumericValue(e.Right)
+			left := KnownPrimitiveType(e.Left)
+			right := KnownPrimitiveType(e.Right)
+			if left == PrimitiveString || right == PrimitiveString {
+				return PrimitiveString
+			}
+			if left == PrimitiveBigInt && right == PrimitiveBigInt {
+				return PrimitiveBigInt
+			}
+			if left != PrimitiveUnknown && left != PrimitiveMixed && left != PrimitiveBigInt &&
+				right != PrimitiveUnknown && right != PrimitiveMixed && right != PrimitiveBigInt {
+				return PrimitiveNumber
+			}
+			return PrimitiveMixed // Can be number or bigint or string (or an exception)
+
+		case BinOpAddAssign:
+			right := KnownPrimitiveType(e.Right)
+			if right == PrimitiveString {
+				return PrimitiveString
+			}
+			return PrimitiveMixed // Can be number or bigint or string (or an exception)
 
 		case
 			BinOpSub, BinOpSubAssign,
 			BinOpMul, BinOpMulAssign,
 			BinOpDiv, BinOpDivAssign,
 			BinOpRem, BinOpRemAssign,
+			BinOpPow, BinOpPowAssign,
 			BinOpBitwiseAnd, BinOpBitwiseAndAssign,
 			BinOpBitwiseOr, BinOpBitwiseOrAssign,
 			BinOpBitwiseXor, BinOpBitwiseXorAssign,
 			BinOpShl, BinOpShlAssign,
 			BinOpShr, BinOpShrAssign,
 			BinOpUShr, BinOpUShrAssign:
-			return true
+			return PrimitiveMixed // Can be number or bigint (or an exception)
 
 		case BinOpAssign, BinOpComma:
-			return IsNumericValue(e.Right)
+			return KnownPrimitiveType(e.Right)
 		}
 	}
 
-	return false
-}
-
-func IsStringValue(a Expr) bool {
-	switch e := a.Data.(type) {
-	case *EInlinedEnum:
-		return IsStringValue(e.Value)
-
-	case *EString:
-		return true
-
-	case *ETemplate:
-		return e.TagOrNil.Data == nil
-
-	case *EIf:
-		return IsStringValue(e.Yes) && IsStringValue(e.No)
-
-	case *EUnary:
-		return e.Op == UnOpTypeof
-
-	case *EBinary:
-		switch e.Op {
-		case BinOpAdd:
-			return IsStringValue(e.Left) || IsStringValue(e.Right)
-
-		case BinOpAssign, BinOpAddAssign, BinOpComma:
-			return IsNumericValue(e.Right)
-		}
-	}
-
-	return false
+	return PrimitiveUnknown
 }
 
 // The goal of this function is to "rotate" the AST if it's possible to use the
@@ -1437,94 +1487,27 @@ const (
 	ImportItemMissing
 )
 
-// Note: the order of values in this struct matters to reduce struct size.
-type Symbol struct {
-	// This is the name that came from the parser. Printed names may be renamed
-	// during minification or to avoid name collisions. Do not use the original
-	// name during printing.
-	OriginalName string
-
-	// This is used for symbols that represent items in the import clause of an
-	// ES6 import statement. These should always be referenced by EImportIdentifier
-	// instead of an EIdentifier. When this is present, the expression should
-	// be printed as a property access off the namespace instead of as a bare
-	// identifier.
-	//
-	// For correctness, this must be stored on the symbol instead of indirectly
-	// associated with the Ref for the symbol somehow. In ES6 "flat bundling"
-	// mode, re-exported symbols are collapsed using MergeSymbols() and renamed
-	// symbols from other files that end up at this symbol must be able to tell
-	// if it has a namespace alias.
-	NamespaceAlias *NamespaceAlias
-
-	// Used by the parser for single pass parsing. Symbols that have been merged
-	// form a linked-list where the last link is the symbol to use. This link is
-	// an invalid ref if it's the last link. If this isn't invalid, you need to
-	// FollowSymbols to get the real one.
-	Link Ref
-
-	// An estimate of the number of uses of this symbol. This is used to detect
-	// whether a symbol is used or not. For example, TypeScript imports that are
-	// unused must be removed because they are probably type-only imports. This
-	// is an estimate and may not be completely accurate due to oversights in the
-	// code. But it should always be non-zero when the symbol is used.
-	UseCountEstimate uint32
-
-	// This is for generating cross-chunk imports and exports for code splitting.
-	ChunkIndex ast.Index32
-
-	// This is used for minification. Symbols that are declared in sibling scopes
-	// can share a name. A good heuristic (from Google Closure Compiler) is to
-	// assign names to symbols from sibling scopes in declaration order. That way
-	// local variable names are reused in each global function like this, which
-	// improves gzip compression:
-	//
-	//   function x(a, b) { ... }
-	//   function y(a, b, c) { ... }
-	//
-	// The parser fills this in for symbols inside nested scopes. There are three
-	// slot namespaces: regular symbols, label symbols, and private symbols.
-	NestedScopeSlot ast.Index32
-
-	Kind SymbolKind
+type SymbolFlags uint8
 
+const (
 	// Certain symbols must not be renamed or minified. For example, the
 	// "arguments" variable is declared by the runtime for every function.
 	// Renaming can also break any identifier used inside a "with" statement.
-	MustNotBeRenamed bool
+	MustNotBeRenamed SymbolFlags = 1 << iota
 
 	// In React's version of JSX, lower-case names are strings while upper-case
 	// names are identifiers. If we are preserving JSX syntax (i.e. not
 	// transforming it), then we need to be careful to name the identifiers
 	// something with a capital letter so further JSX processing doesn't treat
 	// them as strings instead.
-	MustStartWithCapitalLetterForJSX bool
+	MustStartWithCapitalLetterForJSX
 
 	// If true, this symbol is the target of a "__name" helper function call.
 	// This call is special because it deliberately doesn't count as a use
 	// of the symbol (otherwise keeping names would disable tree shaking)
 	// so "UseCountEstimate" is not incremented. This flag helps us know to
 	// avoid optimizing this symbol when "UseCountEstimate" is 1 in this case.
-	DidKeepName bool
-
-	// We automatically generate import items for property accesses off of
-	// namespace imports. This lets us remove the expensive namespace imports
-	// while bundling in many cases, replacing them with a cheap import item
-	// instead:
-	//
-	//   import * as ns from 'path'
-	//   ns.foo()
-	//
-	// That can often be replaced by this, which avoids needing the namespace:
-	//
-	//   import {foo} from 'path'
-	//   foo()
-	//
-	// However, if the import is actually missing then we don't want to report a
-	// compile-time error like we do for real import items. This status lets us
-	// avoid this. We also need to be able to replace such import items with
-	// undefined, which this status is also used for.
-	ImportItemStatus ImportItemStatus
+	DidKeepName
 
 	// Sometimes we lower private symbols even if they are supported. For example,
 	// consider the following TypeScript code:
@@ -1577,7 +1560,105 @@ type Symbol struct {
 	//   };
 	//   Foo.#foo = Foo;
 	//
-	PrivateSymbolMustBeLowered bool
+	PrivateSymbolMustBeLowered
+
+	// This is used to remove the all but the last function re-declaration if a
+	// function is re-declared multiple times like this:
+	//
+	//   function foo() { console.log(1) }
+	//   function foo() { console.log(2) }
+	//
+	RemoveOverwrittenFunctionDeclaration
+)
+
+func (flags SymbolFlags) Has(flag SymbolFlags) bool {
+	return (flags & flag) != 0
+}
+
+// Note: the order of values in this struct matters to reduce struct size.
+type Symbol struct {
+	// This is the name that came from the parser. Printed names may be renamed
+	// during minification or to avoid name collisions. Do not use the original
+	// name during printing.
+	OriginalName string
+
+	// This is used for symbols that represent items in the import clause of an
+	// ES6 import statement. These should always be referenced by EImportIdentifier
+	// instead of an EIdentifier. When this is present, the expression should
+	// be printed as a property access off the namespace instead of as a bare
+	// identifier.
+	//
+	// For correctness, this must be stored on the symbol instead of indirectly
+	// associated with the Ref for the symbol somehow. In ES6 "flat bundling"
+	// mode, re-exported symbols are collapsed using MergeSymbols() and renamed
+	// symbols from other files that end up at this symbol must be able to tell
+	// if it has a namespace alias.
+	NamespaceAlias *NamespaceAlias
+
+	// Used by the parser for single pass parsing. Symbols that have been merged
+	// form a linked-list where the last link is the symbol to use. This link is
+	// an invalid ref if it's the last link. If this isn't invalid, you need to
+	// FollowSymbols to get the real one.
+	Link Ref
+
+	// An estimate of the number of uses of this symbol. This is used to detect
+	// whether a symbol is used or not. For example, TypeScript imports that are
+	// unused must be removed because they are probably type-only imports. This
+	// is an estimate and may not be completely accurate due to oversights in the
+	// code. But it should always be non-zero when the symbol is used.
+	UseCountEstimate uint32
+
+	// This is for generating cross-chunk imports and exports for code splitting.
+	ChunkIndex ast.Index32
+
+	// This is used for minification. Symbols that are declared in sibling scopes
+	// can share a name. A good heuristic (from Google Closure Compiler) is to
+	// assign names to symbols from sibling scopes in declaration order. That way
+	// local variable names are reused in each global function like this, which
+	// improves gzip compression:
+	//
+	//   function x(a, b) { ... }
+	//   function y(a, b, c) { ... }
+	//
+	// The parser fills this in for symbols inside nested scopes. There are three
+	// slot namespaces: regular symbols, label symbols, and private symbols.
+	NestedScopeSlot ast.Index32
+
+	Kind SymbolKind
+
+	// We automatically generate import items for property accesses off of
+	// namespace imports. This lets us remove the expensive namespace imports
+	// while bundling in many cases, replacing them with a cheap import item
+	// instead:
+	//
+	//   import * as ns from 'path'
+	//   ns.foo()
+	//
+	// That can often be replaced by this, which avoids needing the namespace:
+	//
+	//   import {foo} from 'path'
+	//   foo()
+	//
+	// However, if the import is actually missing then we don't want to report a
+	// compile-time error like we do for real import items. This status lets us
+	// avoid this. We also need to be able to replace such import items with
+	// undefined, which this status is also used for.
+	ImportItemStatus ImportItemStatus
+
+	// Boolean values should all be flags instead to save space
+	Flags SymbolFlags
+}
+
+// You should call "MergeSymbols" instead of calling this directly
+func (newSymbol *Symbol) MergeContentsWith(oldSymbol *Symbol) {
+	newSymbol.UseCountEstimate += oldSymbol.UseCountEstimate
+	if oldSymbol.Flags.Has(MustNotBeRenamed) {
+		newSymbol.OriginalName = oldSymbol.OriginalName
+		newSymbol.Flags |= MustNotBeRenamed
+	}
+	if oldSymbol.Flags.Has(MustStartWithCapitalLetterForJSX) {
+		newSymbol.Flags |= MustStartWithCapitalLetterForJSX
+	}
 }
 
 type SlotNamespace uint8
@@ -1590,7 +1671,7 @@ const (
 )
 
 func (s *Symbol) SlotNamespace() SlotNamespace {
-	if s.Kind == SymbolUnbound || s.MustNotBeRenamed {
+	if s.Kind == SymbolUnbound || s.Flags.Has(MustNotBeRenamed) {
 		return SlotMustNotBeRenamed
 	}
 	if s.Kind.IsPrivate() {
@@ -1933,6 +2014,10 @@ type AST struct {
 	ModuleRef   Ref
 	WrapperRef  Ref
 
+	// This contains all top-level exported TypeScript enum constants. It exists
+	// to enable cross-module inlining of constant enums.
+	TSEnums map[Ref]map[string]TSEnumValue
+
 	// These are stored at the AST level instead of on individual AST nodes so
 	// they can be manipulated efficiently without a full AST traversal
 	ImportRecords []ast.ImportRecord
@@ -1954,6 +2039,11 @@ type AST struct {
 	SourceMapComment logger.Span
 }
 
+type TSEnumValue struct {
+	String []uint16 // Use this if it's not nil
+	Number float64  // Use this if "String" is nil
+}
+
 // This is a histogram of character frequencies for minification
 type CharFreq [64]int32
 
@@ -2183,14 +2273,7 @@ func MergeSymbols(symbols SymbolMap, old Ref, new Ref) Ref {
 	}
 
 	oldSymbol.Link = new
-	newSymbol.UseCountEstimate += oldSymbol.UseCountEstimate
-	if oldSymbol.MustNotBeRenamed {
-		newSymbol.OriginalName = oldSymbol.OriginalName
-		newSymbol.MustNotBeRenamed = true
-	}
-	if oldSymbol.MustStartWithCapitalLetterForJSX {
-		newSymbol.MustStartWithCapitalLetterForJSX = true
-	}
+	newSymbol.MergeContentsWith(oldSymbol)
 	return new
 }
 
